"""
定义所有消息类型
"""

__all__ = [
    'MessageSegment',
    'MessageSegmentReceive',
    'MessageSegmentSend',
    'Message',
]


from typing import Literal
from .exceptions import DataFormatError


class MessageSegment(dict):
    def __init__(self, data: dict):
        """
        从dict构建MessageSegment
        """
        if isinstance(data, MessageSegment):
            self = data
        else:
            super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    def __add__(self, other) -> 'Message':
        """
        合并Message
        """
        return Message(self, other)

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'redbag', 'forward', 'share', 'music', 'poke', 'gift', 'node', 'cardimage', 'tts']:
        """消息段类型"""
        return self['type']

    @property
    def data(self) -> dict | None:
        """数据"""
        return self.get('data')


class MessageSegmentReceive(MessageSegment):
    def __init__(self, data: dict):
        """
        从dict构建MessageSegmentReceive
        """
        super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'redbag', 'forward']:
        """消息段类型"""
        return self['type']


class MessageSegmentSend(MessageSegment):
    def __init__(self, data: dict):
        """
        从dict构建MessageSegmentSend
        """
        super().__init__(data)
        try:
            _ = self.type
        except KeyError as e:
            raise DataFormatError(e)

    @property
    def type(self) -> Literal['text', 'face', 'record', 'video', 'at', 'share', 'image', 'reply', 'xml', 'json', 'share', 'music', 'poke', 'gift', 'node', 'cardimage', 'tts']:
        """消息段类型"""
        return self['type']

    @staticmethod
    def build(type: str, **data):  # TODO 下面的CQ码构造并没有写全，参考https://docs.go-cqhttp.org/cqcode/
        """
        构建MessageSegmentSend
        """
        return MessageSegmentSend({
            'type': type,
            'data': data
        })

    @staticmethod
    def text(text: str):
        """纯文本
        - `text`: 文本
        """
        return MessageSegmentSend.build('text', text=text)

    @staticmethod
    def face(id: int):
        """QQ 表情
        - `id`: QQ 表情 ID，参考https://github.com/kyubotics/coolq-http-api/wiki/%E8%A1%A8%E6%83%85-CQ-%E7%A0%81-ID-%E8%A1%A8
        """
        return MessageSegmentSend.build('face', id=id)

    @staticmethod
    def at(qq: int | Literal['all'], name: str = ''):
        """@某人
        - `qq`: @的 QQ 号, all 表示全体成员
        - `name`: 当在群中找不到此QQ号的名称时才会生效
        """
        if name == '':
            return MessageSegmentSend.build('at', qq=qq)
        else:
            return MessageSegmentSend.build('at', qq=qq, name=name)


class Message(list[MessageSegment]):
    def __init__(self, *data):
        """
        从dict构建Message
        """
        if not data:
            raise DataFormatError('data is empty')
        super().__init__()
        for segment in data:
            if isinstance(segment, Message):
                self.extend(segment)
            elif isinstance(segment, MessageSegment):
                self.append(segment)
            elif isinstance(segment, (list, tuple)):
                self.extend(MessageSegment(s) for s in segment)
            elif isinstance(segment, dict):
                self.append(MessageSegment(segment))
            else:
                raise DataFormatError(
                    f'{segment} is not a MessageSegment or Message')

    def __add__(self, other):
        """合并Message"""
        return Message(self, other)



"""
定义交互器，用于和OneBot实现交互
包括Http和正向WebSocket通讯协议
包括调用Api和接收Event

InteractorApi：
    将信息发出去以后，异步等待返回，消息返回后由InteractorEvent处理，最后返回 ApiResponse 对象
InteractorEvent：
    使用event.py将接收到的Event(dict)，转换为Event对象，并且交由自己的handlers处理
"""

from abc import ABC, abstractmethod
import asyncio
from typing import Any, Callable

import aiohttp

try:
    import ujson as json
except ImportError:
    import json

from .type import ApiResponse
from .utils import EchoHandler
from .event import Event
from .exceptions import WebSocketNotConnected

__all__ = [
    'Interactor',
    'InteractorApi',
    'InteractorEvent',
    'InteractorHttpApi',
    'InteractorHttpEvent',
    'InteractorWebSocket',
]


class Interactor:
    def __init__(
        self, url: str, 
        *,
        access_token: str = '',
    ):

        self.url = url
        self.access_token = access_token
        self.session = aiohttp.ClientSession(json_serialize=json.dumps)


class InteractorApi(Interactor, ABC):
    def __init__(
        self, url: str, 
        *,
        access_token: str = '',
        timeout: float = 10
    ):
        super().__init__(url, access_token=access_token)
        # 默认超时时长为10秒
        self._timeout = timeout

    @abstractmethod
    async def call(self, action: str, **params) -> 'ApiResponse':
        """请求OneBot实现，并获取返回值"""


class InteractorEvent(Interactor, ABC):
    def __init__(
        self, url: str, 
        *,
        access_token: str = '',
    ):
        super().__init__(url, access_token=access_token)
        self.handlers: list[Callable[['Event'], Any]] = []

    @abstractmethod
    async def _receiver(self):
        """接收Event"""

    async def receive(self, event_loop: asyncio.AbstractEventLoop):
        event_loop.create_task(self._receiver())


class InteractorHttpApi(InteractorApi):
    async def call(self, action: str, **params) -> 'ApiResponse':
        data = {
            'action': action,
            'params': params,
        }
        async with self.session.post(self.url, data=data, timeout=self._timeout) as msg:
            resp = await msg.json()
        resp = ApiResponse(resp)
        return resp


class InteractorHttpEvent(InteractorApi):
    """暂未实现"""


class InteractorWebSocket(InteractorApi, InteractorEvent):
    def __init__(
        self, url: str, 
        *,
        access_token: str = '',
        timeout: float = 10
    ):
        super().__init__(url, access_token=access_token, timeout=timeout)
        self.conn = False

    async def connect(self, event_loop: asyncio.AbstractEventLoop):
        self.ws_session = await self.session.ws_connect(self.url)
        self.conn = True
        await self.receive(event_loop)

    async def close(self):
        if not self.conn:
            raise WebSocketNotConnected
        await self.ws_session.close()
        self.conn = False

    async def call(self, action: str, **params) -> 'ApiResponse':
        if not self.conn:
            raise WebSocketNotConnected

        echo = EchoHandler.next()
        EchoHandler.add_fut(echo)
        data = {
            'action': action,
            'params': params,
            'echo': echo,
        }
        await self.ws_session.send_json(data)

        resp = await EchoHandler.wait(echo, self._timeout)
        resp = ApiResponse(resp)
        return resp

    async def _receiver(self):
        while True:
            data = await self.ws_session.receive_json()
            if data.get('echo') is None:
                # 正常上报
                event = Event(data)
                asyncio.gather(*[handler(event) for handler in self.handlers])
            else:
                # echo上报
                EchoHandler.set(data)



"""
定义与Bot相关的东西
Bot需要将自己（或某个method）设置为Interactor的handler，一个Bot可以绑定多个Interactor
Bot提供on（及其附属）装饰器，接收优先级，并注册handler
在Bot接收到Event时，按照优先级依次使用handler处理消息，根据handler的参数类型来输入参数
"""


from typing import Any, Callable, get_type_hints

from .utils import event_handler_sort, EventHandler
from .event import Event
from .exceptions import StopEventProcessing

__all__ = [
    'Bot',
]


class Bot:
    def __init__(self):
        self.handlers: list[EventHandler] = []

    async def __call__(self, event: 'Event'):
        """
        根据优先级依次执行处理函数
        """
        sorted_handlers = event_handler_sort(self.handlers)
        try:
            for handler in sorted_handlers:
                if handler.assert_condition(event):
                    await handler(event)
                else:
                    # TODO 这里应该弄一个warn的log
                    pass
        except StopEventProcessing:
            return

    def on(
        self,
        *,
        priority: int = 0,
        **condition
    ) -> Callable:
        """注册事件处理函数
        - `priority`: 优先级：正数中，越小优先级越高；0为无优先级，其次执行；负数最后执行；如果优先级冲突，则按照添加顺序执行
        - `**condition`: 事件满足条件才会触发
        """
        def decorator(func: Callable[['Event'], Any]):
            type_hints = get_type_hints(func)
            if 'return' in type_hints:
                type_hints.pop('return')
            if len(type_hints.items()) != 1:
                raise TypeError('Handler function must have only one arg')
            self.handlers.append(EventHandler(func, priority=priority, condition=condition))
            return func

        return decorator
